/*
   assign2.c prints a maze and calculates how many
   dead-ends and branches there are, if it's a
   valid maze.

   author: Beno Gorancic
   date  : 26/05/2006
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define TRUE 1
#define FALSE 0

struct Columns
{
   char value;
   struct Columns *next;
};

struct Rows {
   struct Columns *firstColumn;
   struct Rows *next;
};

typedef struct Columns column;
typedef struct Rows row;

FILE* openFile(char *filename);
row *addElements(char* filename);
void printElements(row *rhead);
void calculateMetrics(row *rhead);
void freeList(row *rhead);
int neighbourValue(row *rhead, int rowCount, int colCount);
int isInMaze(row *rhead, char letter);
int isValid(row *rhead);

int main(int argc, char *argv[])
{
   row* myList;

   if (argc != 2)
   {
      printf("Must supply 1 argument to this program\n");
      exit(0);
   }

   myList = addElements(argv[1]);

   if (!(isInMaze(myList, 's')))
      printf("Error - start declared outside of maze\n");

   else if (!(isInMaze(myList, 'f')))
      printf("Error - finish declared outside of maze\n");

   else if(!(isValid(myList)))
      exit(0);

   else
   {
      printElements(myList);
      calculateMetrics(myList);
   }

   freeList(myList);
   exit(0);
}

FILE* openFile(char *filename)
{
   /* open a FILE stream to filename with given mode */

   FILE *file = fopen(filename, "r");

   if (file == NULL)
   {
      printf("ERROR - Unable to open file %s\n", filename);
      exit(0);
   }

   return file;
}

row *addElements(char* filename)
{
   /* adds chars into the linked list. Every line in
      the text file is a new row in the linked list */

   row *rhead = NULL, *rCurrent, *rTemp;
   column *cCurrent, *cTemp;

   int val;
 
   FILE *fin = openFile(filename);

   rCurrent = malloc(sizeof(row));
   rhead = rCurrent;
   rhead->next = NULL;
   cCurrent = NULL;

   /* get charater until the end of file is reached */
   
   while((val = fgetc(fin)) != EOF)
   {
      /* if it's not a new line add character to the column
          else create a new row */

      if(val != '\n')
      {
         cTemp = malloc(sizeof(column));
         
         if(cCurrent == NULL)
         {
            rCurrent->firstColumn = cTemp;
            cCurrent = cTemp;
            cCurrent->next = NULL;
         }
         else
         {
            cCurrent->next = cTemp;
            cCurrent = cTemp;
            cCurrent->next = NULL;
         }
         cCurrent->value = val;
         cCurrent->next = NULL;
      }
      else
      {
         rTemp = malloc(sizeof(row));
         rCurrent->next = rTemp;
         rCurrent = rTemp;
         rCurrent->next = NULL;
         rCurrent->firstColumn = NULL;
         cCurrent = NULL;
      }
   }

   fclose(fin);
   return rhead;
}

void printElements(row *rhead)
{
   /* prints the chars stored in the linked list */

   row *rCurrent = rhead;
   column *cCurrent;

   /* repeat while more rows avilable */

   while(rCurrent)
   {
      cCurrent = rCurrent->firstColumn;

      /* while more chars available, print char */

      while(cCurrent)
      {
         printf("%c", cCurrent->value);
         cCurrent = cCurrent->next;
      }

      printf("\n");
      rCurrent = rCurrent->next;
   }
}

void calculateMetrics(row *rhead)
{
   /* calculates the number of branches and dead-ends in the maze */

   row *rCurrent = rhead, *rCheck = rhead;
   column *cCurrent;

   char top, bottom, left, right, ch;
   int rowCount = 0, colCount = 0, spaces = 0;
   int brachCount = 0, deadEndCount = 0;

   rCurrent = rCurrent->next;

   while(rCurrent)
   {
      rowCount++;
      cCurrent = rCurrent->firstColumn;

      /* increment while more chars available & char is not equal to '#' */
      while(cCurrent && (cCurrent->value != '#'))
      {
         colCount++;
         cCurrent = cCurrent->next;
      }

      /* while more columns available after the 1st '#' (above) is found */
      while(cCurrent)
      {
         ch = cCurrent->value;

         /* if char is not a '#' find what the surrounding char is */
         if(ch != '#')
         {
            top = neighbourValue(rCheck, rowCount-1, colCount);
            bottom = neighbourValue(rCheck, rowCount+1, colCount);
            left = neighbourValue(rCheck, rowCount, colCount-1);
            right = neighbourValue(rCheck, rowCount, colCount+1);

            /* add number of spaces found in either direction */
            if(top == ' ') spaces++;
            if(bottom == ' ') spaces++;
            if(left == ' ') spaces++;
            if(right == ' ') spaces++;

            /* depending on the number of spaces found & if a surrouding char
               is an s or f, the number of branches & dead-ends is counted */

            if((spaces == 3) || (ch == 's' && spaces == 2)) brachCount++;

            if((spaces == 4) || (ch == 's' && spaces == 3)) brachCount +=2;

            if(spaces == 1 && !strchr("fs", top) && !strchr("fs", bottom) &&
               !strchr("fs", left) && !strchr("fs", right) &&
               !strchr("fs", ch))
                  deadEndCount++;

            spaces = 0;
         }
         colCount++;
         cCurrent = cCurrent->next;
      }
      colCount=0;
      rCurrent = rCurrent->next;
   }
   printf("\nNumber of branches in maze is  %d\n", brachCount);
   printf("Number of dead-ends in maze is %d\n", deadEndCount);
}

void freeList(row *rhead)
{
   /* frees memory it previously allocated with malloc */

   row *rCurrent = rhead, *rTemp;
   column *cCurrent, *cTemp;

   /* repeat while more rows avilable */

   while(rCurrent)
   {
      cCurrent = rCurrent->firstColumn;

      /* repeat while more columns (chars in row) available */

      while(cCurrent)
      {
         cTemp = cCurrent->next;
         free(cCurrent);
         cCurrent = cTemp;
      }

      rTemp = rCurrent->next;
      free(rCurrent);
      rCurrent = rTemp;
   }
}

int neighbourValue(row *rhead, int rowCount, int colCount)
{
   /* given the position (row number & column number) of a char,
      the value next to it is found by moving through the list
      starting at the top of it */

   row *rCurrent = rhead;
   column *cCurrent;

   int cCount = 0, rCount = 0;

   /* move down the rows until the given row is reached */

   while(rCount != (rowCount))
   {
      rCurrent = rCurrent->next;
      rCount++;
   }

   cCurrent = rCurrent->firstColumn;

   /* having moved to the correct row, move through the 
      columns (chars) until the give column is reached*/

   while((cCount != colCount) && (cCurrent != NULL))
   {
      cCurrent = cCurrent->next;
      cCount++;
   }

   /* having reached the particular point, return it's value
      if it's not null, otherwise return a 0 */

   if (cCurrent != NULL) return cCurrent->value;

   return 0;
}

int isInMaze(row *rhead, char letter)
{
   /* checks to see if a char (letter) is inside the maze */

   row *rCurrent = rhead, *rTopCheck, *rBottomCheck;
   column *cCurrent, *cLeftCheck;

   char ch_top, ch_bottom;

   int rowCount = 0, colCount = 0;
   int leftSpaces = 0, topSpaces = 0, bottomSpaces = 0;
   int bottomRows = 0, i;

   while(rCurrent)
   {
      cCurrent = rCurrent->firstColumn;
      cLeftCheck = rCurrent->firstColumn;
      rTopCheck = rhead;
      rBottomCheck = rCurrent->next;

      while(cCurrent)
      {
         /* if current char equals the letter to be checked */
         if(cCurrent->value == letter)
         {
            /* count spaces to the left of letter */
            while(cLeftCheck->value != cCurrent->value)
            {
               if(cLeftCheck->value == ' ') leftSpaces++;
               cLeftCheck = cLeftCheck->next;
            }

            /* count spaces above letter */
            for(i=0; i < rowCount; i++)
            {
               ch_top = neighbourValue(rTopCheck, i, colCount);
               if (ch_top == ' ') topSpaces++;
            }

            /* count spaces bellow letter */
            while(rBottomCheck)
            {
               ch_bottom = neighbourValue(rBottomCheck, 0, colCount);
               if (ch_bottom == ' ') bottomSpaces++;
               if (ch_bottom != 0) bottomRows++;
               rBottomCheck = rBottomCheck->next;
            }

            /* if in either direction, number of spaces equals the number
               of available chars in that direction up to letter being
               checked, means only spaces present. Letter is outside the
               maze. Test fails & rest of the program isn't executed */

            if((leftSpaces == colCount) || (topSpaces == rowCount) || 
               (bottomSpaces == bottomRows))
                  return FALSE;
         }
         colCount++;
         cCurrent = cCurrent->next;
      }
      colCount=0;
      rowCount++;
      rCurrent = rCurrent->next;
   }

   return TRUE;
}

int isValid(row *rhead)
{
   /* checks to see if only the following chars are present
      in the maze: #'s, spaces, one 's' and one 'f' */

   row *rCurrent = rhead;
   column *cCurrent;
   char ch;
   int numOfS = 0, numOfF = 0;

   /* repeat while more rows avilable */
   while (rCurrent)
   {
      cCurrent = rCurrent->firstColumn;

      /* repeat while more columns (chars in row) available */
      while(cCurrent)
      {
         ch = cCurrent->value;

         /* If char other than #, space, s or f present, return false */

         if(!(ch == '#' || ch == ' '|| ch == 's'|| ch == 'f'))
         {
            printf("Error - invalid maze character\n");
            return FALSE;
         }

         /* otherwise count the number of 's' and 'f' present */

         else
         {
            if(ch == 's') numOfS++;
            if(ch == 'f') numOfF++;
         }
         cCurrent = cCurrent->next;
      }
      rCurrent = rCurrent->next;
   }

   /* if more/less 's' or 'f' found, display message and return false */

   if(numOfS > 1)
   {
      printf("Error - multiple start found in maze\n");
      return FALSE;
   }
   
   if(numOfF > 1)
   {
      printf("Error - multiple finish found in maze\n");
      return FALSE;
   }

   if (numOfS < 1)
   {
      printf("Error - no start found in maze\n");
      return FALSE;
   }
   
   if (numOfF < 1)
   {
      printf("Error - no finish found in maze\n");
      return FALSE;
   }

   return TRUE;
}
